//! Implements the Lua API for the UTXO contract.
use crate::vm::{Error, Instance, Storage, Value, Variable};
use futures::executor::block_on;
use mlua::{Compiler, Lua, Table, Value as luaValue, Variadic};
use std::{collections::HashMap, sync::Arc};

impl TryFrom<&luaValue<'_>> for Value {
    type Error = String;

    fn try_from(value: &luaValue<'_>) -> Result<Self, Self::Error> {
        match value {
            luaValue::Nil => Ok(Value::Nil),
            luaValue::Boolean(b) => Ok(Value::Boolean(*b)),
            luaValue::Integer(i) => Ok(Value::Integer((*i).into())),
            luaValue::Number(n) => Ok(Value::Number(*n)),
            luaValue::String(s) => Ok(Value::String(s.to_str().unwrap().to_owned())),
            luaValue::Table(t) => {
                let mut map = HashMap::new();
                let mut iter = t.clone().pairs::<String, luaValue>().enumerate();
                let mut is_vector = true;
                while let Some((id, Ok((k, v)))) = iter.next() {
                    if id.checked_add(1) != k.parse().ok() {
                        is_vector = false;
                    }
                    map.insert(k, v.as_ref().try_into()?);
                }

                Ok(if is_vector {
                    let mut values = map
                        .into_iter()
                        .map(|(k, v)| k.parse().map(|k| (k, v)))
                        .collect::<Result<Vec<(usize, Value)>, _>>()
                        .unwrap();

                    values.sort_by(|(a, _), (b, _)| a.cmp(b));

                    Value::Vector(values.into_iter().map(|(_, v)| v).collect())
                } else {
                    Value::HashMap(map)
                })
            }
            x => Err(format!("Invalid type: {:?}", x)),
        }
    }
}

/// Implements the Lua Virtual machine support for the ledger database
pub struct LuaVM<X: Storage + 'static> {
    state: Arc<X>,
    opcodes: Vec<u8>,
}

impl<X: Storage + 'static> LuaVM<X> {
    #[inline]
    fn var_value_to_lua_val(lua: &Lua, value: Value) -> mlua::Result<luaValue> {
        match value {
            Value::Nil => Ok(luaValue::Nil),
            Value::Boolean(b) => Ok(luaValue::Boolean(b)),
            Value::Integer(i) => Ok(luaValue::Integer(i.try_into().unwrap())),
            Value::Number(n) => Ok(luaValue::Number(n)),
            Value::String(s) => Ok(luaValue::String(lua.create_string(&s)?)),
            Value::HashMap(map) => {
                let table = lua.create_table()?;
                for (k, v) in map {
                    table.set(k, Self::var_value_to_lua_val(lua, v)?)?;
                }
                Ok(luaValue::Table(table))
            }
            Value::ErrorType(e) => Err(mlua::Error::RuntimeError(e.to_string())),
            _ => Err(mlua::Error::RuntimeError("Invalid type".into())),
        }
    }

    #[inline]
    fn crate_meta_variables(lua: &Lua, meta_variable: Variable<'static>) -> mlua::Result<()> {
        let getter = lua.create_function(move |_, (_, key): (Table, String)| {
            println!("read only variable: {:?}.{}", meta_variable, key);
            Ok(1)
        })?;

        let setter =
            lua.create_function(move |_, (_, key, value): (Table, String, luaValue)| {
                Result::<String, _>::Err(mlua::Error::RuntimeError(format!(
                    "read only variable: {:?}.{} = {:?}",
                    meta_variable, key, value
                )))
            })?;

        let table = lua.create_table()?;
        let meta_table = lua.create_table()?;
        meta_table.set("__index", getter)?;
        meta_table.set("__newindex", setter)?;
        table.set_metatable(Some(meta_table));
        lua.globals().raw_set(meta_variable.name(), table)?;
        Ok(())
    }

    #[inline]
    fn inject_dynamic_global_state(
        lua: &Lua,
        storage: Arc<X>,
        program_name: String,
        instance: usize,
    ) -> mlua::Result<Option<Table>> {
        lua.set_app_data((program_name, storage));

        let getter = lua.create_function(move |lua, (global, key): (Table, String)| {
            match global.raw_get::<_, luaValue>(key.clone())? {
                luaValue::Nil => (),
                local_value => return Ok(local_value),
            };
            let (program_name, storage) = lua
                .app_data_ref::<(String, Arc<X>)>()
                .ok_or(mlua::Error::MismatchedRegistryKey)?
                .clone();

            let key = Variable::new(&program_name, &key);
            // LuaVM does not like being de-scheduled to run async tasks, any async task must
            // be executed on the same thread
            let value = block_on(async move { storage.get(instance, key).await });
            Self::var_value_to_lua_val(lua, value)
        })?;
        let setter = lua.create_function(
            move |lua, (global, key, value): (Table, String, luaValue)| {
                let (program_name, storage) = lua
                    .app_data_ref::<(String, Arc<X>)>()
                    .ok_or(mlua::Error::MismatchedRegistryKey)?
                    .clone();
                let value: Value = if let Ok(value) = value.as_ref().try_into() {
                    value
                } else {
                    return global.raw_set(key, value);
                };
                let key = Variable::new(&program_name, &key);
                // LuaVM does not like being de-scheduled to run async tasks, any async task must
                // be executed on the same thread
                block_on(async move {
                    storage.set(instance, key, value).await;
                    Ok(())
                })
            },
        )?;

        let metatable = lua.create_table()?;
        metatable.raw_set("__index", getter)?;
        metatable.raw_set("__newindex", setter)?;

        Ok(Some(metatable))
    }
}

#[async_trait::async_trait]
impl<X: Storage + 'static> Instance<X> for LuaVM<X> {
    async fn new(state: Arc<X>, code: &str) -> Result<Self, Error>
    where
        Self: Sized,
    {
        Ok(Self {
            state,
            opcodes: Compiler::new().compile(code),
        })
    }

    async fn run(
        &self,
        execution_id: usize,
        program_name: &str,
        args: Vec<Value>,
    ) -> Result<Value, Error> {
        let lua = Lua::new();
        let globals = lua.globals();

        let require = lua
            .create_function(|_, (_,): (String,)| -> mlua::Result<()> {
                Err(mlua::Error::RuntimeError("require is not allowed".into()))
            })
            .map_err(|e| Error::Runtime(e.to_string()))?;

        globals.set_metatable(
            Self::inject_dynamic_global_state(
                &lua,
                self.state.clone(),
                program_name.to_owned(),
                execution_id,
            )
            .map_err(|e| Error::Runtime(e.to_string()))?,
        );

        for val in [
            Variable::Accounts,
            Variable::Balances,
            Variable::Transactions,
            Variable::Payments,
        ] {
            Self::crate_meta_variables(&lua, val).map_err(|e| Error::Runtime(e.to_string()))?;
        }

        // remove external require
        globals
            .set("require", require)
            .map_err(|e| Error::Runtime(e.to_string()))?;
        drop(globals);

        let args = Variadic::from_iter(
            args.into_iter()
                .map(|x| Self::var_value_to_lua_val(&lua, x))
                .collect::<Result<Vec<_>, _>>()
                .map_err(|e| Error::Runtime(e.to_string()))?,
        );

        let result: luaValue = lua
            .load(&self.opcodes)
            .call(args)
            .map_err(|e| Error::Runtime(e.to_string()))?;

        result.as_ref().try_into().map_err(Error::Runtime)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use std::sync::Arc;

    #[tokio::test]
    async fn test() {
        let storage = Arc::new(crate::vm::memory::Storage::default());
        let vm = LuaVM::new(
            storage.clone(),
            r#"
            subscriptions = {}
            local arg = {...}
            function update_state()
                set = "foo " .. arg[1]
                return arg[1]
            end

            return update_state()
            "#,
        )
        .await
        .unwrap();

        let v = vm.run(0, "test", vec![Value::Integer(2)]).await.unwrap();
        assert_eq!(v.as_integer(), Some(2));

        let vm = LuaVM::new(
            storage.clone(),
            r#"
            return set
            "#,
        )
        .await
        .unwrap();

        let v = vm.run(0, "test", vec![]).await.unwrap();
        assert_eq!(v.as_str(), Some("foo 2"));

        let vm = LuaVM::new(
            storage.clone(),
            r#"
            return balances["foo"]
            "#,
        )
        .await
        .unwrap();
        let v = vm.run(0, "test", vec![]).await.unwrap();
        assert_eq!(v.as_integer(), Some(1));
    }

    #[tokio::test]
    async fn require_not_allowed() {
        let storage = Arc::new(crate::vm::memory::Storage::default());
        let vm = LuaVM::new(
            storage.clone(),
            r#"
            require("foo")
            "#,
        )
        .await
        .unwrap();

        let r = vm.run(0, "test", vec![]).await;
        assert!(r.is_err());
        assert!(r
            .unwrap_err()
            .to_string()
            .find("require is not allowed")
            .is_some());
    }
}
