//! Abstract definition of a virtual machine and their programs to be executed
//! inside the ledger for each transaction
use std::{
    collections::HashMap,
    sync::{atomic::AtomicUsize, Arc},
};
use tokio::sync::RwLock;

mod program;
mod state;

pub use self::{program::Instance, state::*};

#[derive(thiserror::Error, Debug)]
/// Error type for the virtual machine
pub enum Error {
    /// A program is not found
    #[error("Program not found: {0}")]
    ProgramNotFound(String),

    /// A given instance is not valid
    #[error("Instance {1} for program {0} not found")]
    InstanceNotFound(String, usize),

    #[error("Error compiling program: {0}")]
    /// Compilation error
    Compilation(String),

    #[error("Runtime error: {0}")]
    /// Runtime error
    Runtime(String),
}

/// Creates a runtime for the virtual machine
///
/// Each program will be have their own Program and their own storage. Only a
/// few special variables are shared between programs (such as balances,
/// accounts, transactions and payments), all other variables are unique per
/// program.
pub struct Runtime<S, I>
where
    S: Storage + 'static,
    I: Instance<S> + 'static,
{
    programs: Arc<RwLock<HashMap<String, program::Program<S, I>>>>,
    storage: Arc<S>,
    instance: AtomicUsize,
    _phantom: std::marker::PhantomData<S>,
}

impl<X, P> Runtime<X, P>
where
    X: Storage + 'static,
    P: Instance<X> + 'static,
{
    /// Creates a new instance of the virtual machine
    pub fn new(storage: Arc<X>) -> Arc<Self> {
        Arc::new(Self {
            programs: Arc::new(RwLock::new(HashMap::new())),
            instance: AtomicUsize::new(0),
            storage,
            _phantom: std::marker::PhantomData,
        })
    }

    /// Creates or replaces a program
    ///
    /// If there was a previous program with the same name, it will be replaced
    /// and the old instance will be returned
    pub async fn create(&self, name: &str, code: &str) -> Result<(), Error> {
        let mut programs = self.programs.write().await;
        programs.insert(
            name.to_string(),
            program::Program::new(self.storage.clone(), name, code)?,
        );
        Ok(())
    }

    /// Runs a program
    pub async fn run(&self, name: &str, input: Vec<Value>) -> Result<(usize, Value), Error> {
        let execution_id = self
            .instance
            .fetch_add(1, std::sync::atomic::Ordering::SeqCst);
        let programs = self.programs.read().await;
        let program = programs
            .get(name)
            .ok_or(Error::ProgramNotFound(name.to_string()))?;
        Ok((execution_id, program.run(execution_id, input).await?))
    }
}
