use super::{Value, Variable};
use std::collections::HashMap;
use tokio::{
    sync::{Mutex, RwLock},
    time::Duration,
};

#[derive(Debug)]
pub struct Storage {
    storage: RwLock<HashMap<String, Value>>,
    locks: RwLock<HashMap<String, usize>>,
    var_locked_by_instance: Mutex<HashMap<usize, Vec<String>>>,
}

impl Default for Storage {
    fn default() -> Self {
        Self {
            storage: RwLock::new(HashMap::new()),
            locks: RwLock::new(HashMap::new()),
            var_locked_by_instance: Mutex::new(HashMap::new()),
        }
    }
}

#[async_trait::async_trait]
impl super::Storage for Storage {
    async fn lock<'a>(&'a self, instance: usize, var: &'a Variable<'a>) {
        let locks = self.locks.read().await;
        let name = var.name();

        if locks.get(&name).map(|v| *v) == Some(instance) {
            // The variable is already locked by this instance
            return;
        }

        drop(locks);

        loop {
            // wait here while the locked is not null or it is locked by another
            // instance
            let locks = self.locks.read().await;
            let var_lock = locks.get(&name).map(|v| *v);
            if var_lock.is_none() || var_lock == Some(instance) {
                break;
            }
            drop(locks);
            tokio::time::sleep(Duration::from_micros(10)).await;
        }

        loop {
            let mut locks = self.locks.write().await;
            let var_lock = locks.get(&name).map(|v| *v);

            if !var_lock.is_none() {
                if var_lock == Some(instance) {
                    break;
                }
                drop(locks);
                tokio::time::sleep(Duration::from_micros(10)).await;
                continue;
            }

            locks.insert(name.to_owned(), instance);

            let mut vars_by_instance = self.var_locked_by_instance.lock().await;
            if let Some(vars) = vars_by_instance.get_mut(&instance) {
                vars.push(name.to_owned());
            } else {
                vars_by_instance.insert(instance, vec![name.to_owned()]);
            }
        }
    }

    async fn get<'a>(&'a self, instance: usize, var: Variable<'a>) -> Value {
        self.lock(instance, &var).await;
        self.storage
            .read()
            .await
            .get(&var.name())
            .cloned()
            .unwrap_or(Value::Nil)
    }

    async fn set<'a>(&'a self, instance: usize, var: Variable<'a>, value: Value) {
        self.lock(instance, &var).await;
        self.storage.write().await.insert(var.name(), value);
    }

    async fn shutdown(&self, execution_id: usize) {
        let mut vars_by_instance = self.var_locked_by_instance.lock().await;
        let mut locks = self.locks.write().await;

        if let Some(vars) = vars_by_instance.remove(&execution_id) {
            for var in vars {
                if locks.get(&var).map(|v| *v) == Some(execution_id) {
                    locks.remove(&var);
                }
            }
        }
    }
}
