use std::collections::HashMap;

#[cfg(test)]
pub mod memory;

/// The value of the data to store
#[derive(Debug, Clone)]
pub enum Value {
    /// The Lua value `nil`.
    Nil,
    /// The Lua value `true` or `false`.
    Boolean(bool),
    /// Integer number
    Integer(i128),
    /// A floating point number.
    Number(f64),
    /// String
    String(String),
    /// A vector
    Vector(Vec<Value>),
    /// Dynamic value tied to a prefix, each values are retrieved one by one and
    /// the locking is done on each element value
    DynamicHashMap(String),
    /// A
    HashMap(HashMap<String, Value>),
    /// An error
    ErrorType(String),
}

impl Value {
    /// Gets value as a &str if possible
    pub fn as_str(&self) -> Option<&str> {
        match self {
            Self::String(s) => Some(s),
            _ => None,
        }
    }

    /// Gets value as a i128 if possible
    pub fn as_integer(&self) -> Option<i128> {
        match self {
            Self::Integer(i) => Some(*i),
            _ => None,
        }
    }

    /// Is the current value a hashmap. Hashmap are stored in a flat structure
    /// in the storage layer, each value is locked individually instead of the
    /// whole hashmap
    pub fn is_hashmap(&self) -> bool {
        match self {
            Self::HashMap(_) | Self::DynamicHashMap(_) => true,
            Self::Vector(x) => x.is_empty(),
            _ => false,
        }
    }
}

/// Variable name
#[derive(Debug, Clone, Eq, PartialEq, Copy)]
pub enum Variable<'a> {
    /// Balance: This variable is shared between all programs
    Balances,
    /// Accounts: This variable is shared between all programs
    Accounts,
    /// Transactions: This variable is shared between all programs
    Transactions,
    /// Payments: This variable is shared between all programs
    Payments,
    /// Dynamic
    Dynamic(&'a Variable<'a>, &'a str),
    /// Any local variable, inside the namespace of the current program
    Scalar(&'a str, &'a str),
}

impl<'a> Variable<'a> {
    /// Creates a new variable
    ///
    /// The variable is bound to a namespace (the program) name unless it is an
    /// special global / share variable such as balances, accounts, transactions
    /// or payments.
    pub fn new(namespace: &'a str, name: &'a str) -> Self {
        match namespace {
            "balances" => Self::Balances,
            "accounts" => Self::Accounts,
            "transactions" => Self::Transactions,
            "payments" => Self::Payments,
            _ => Self::Scalar(namespace, name),
        }
    }

    /// Whether the current variable is a meta variable (balances, accounts, transactions or payments)
    pub fn is_meta_variable(&self) -> bool {
        matches!(
            self,
            Self::Balances | Self::Accounts | Self::Transactions | Self::Payments
        )
    }

    /// Returns the name of the variable as an String
    pub fn name(&self) -> String {
        match self {
            Self::Balances => "balances".to_owned(),
            Self::Accounts => "accounts".to_owned(),
            Self::Transactions => "transactions".to_owned(),
            Self::Payments => "payments".to_owned(),
            Self::Dynamic(parent, name) => format!("{}.{}", parent.name(), name),
            Self::Scalar(ns, name) => format!("{}.{}", ns, name),
        }
    }
}

#[async_trait::async_trait]
/// The storage trait to implement a storage engine
pub trait Storage: Send + Sync {
    /// Locks a variable for a given execution_id, it will be automatically
    /// released when shutdown(execution_id is being called)
    async fn lock<'a>(&'a self, execution_id: usize, var: &'a Variable<'a>);

    /// Gets a single value from the storage. The returned value should be locked by the execution_id
    async fn get<'a>(&'a self, execution_id: usize, variable: Variable<'a>) -> Value;

    /// Sets a value in the storage. The variable should be locked by the
    /// execution_id before updating
    async fn set<'a>(&'a self, execution_id: usize, var: Variable<'a>, value: Value);

    /// release resources)
    async fn shutdown(&self, execution_id: usize);
}
